"""
The RMAType definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


from typing import Union, Optional

import numpy
from numpy.linalg import norm

from sarpy.io.xml.base import Serializable
from sarpy.io.xml.descriptors import StringEnumDescriptor, FloatDescriptor, \
    BooleanDescriptor, SerializableDescriptor

from .base import DEFAULT_STRICT, FLOAT_FORMAT
from .blocks import XYZType, Poly1DType, Poly2DType
from .utils import _get_center_frequency


class RMRefType(Serializable):
    """
    Range migration reference element of RMA type.
    """

    _fields = ('PosRef', 'VelRef', 'DopConeAngRef')
    _required = _fields
    _numeric_format = {'DopConeAngRef': FLOAT_FORMAT, }
    # descriptors
    PosRef = SerializableDescriptor(
        'PosRef', XYZType, _required, strict=DEFAULT_STRICT,
        docstring='Platform reference position in ECF coordinates used to establish '
                  'the reference slant plane.')  # type: XYZType
    VelRef = SerializableDescriptor(
        'VelRef', XYZType, _required, strict=DEFAULT_STRICT,
        docstring='Platform reference velocity vector in ECF coordinates used to establish '
                  'the reference slant plane.')  # type: XYZType
    DopConeAngRef = FloatDescriptor(
        'DopConeAngRef', _required, strict=DEFAULT_STRICT,
        docstring='Reference Doppler Cone Angle in degrees.')  # type: float

    def __init__(
            self,
            PosRef: Union[XYZType, numpy.ndarray, list, tuple] = None,
            VelRef: Union[XYZType, numpy.ndarray, list, tuple] = None,
            DopConeAngRef: float = None,
            **kwargs):
        """

        Parameters
        ----------
        PosRef : XYZType|numpy.ndarray|list|tuple
        VelRef : XYZType|numpy.ndarray|list|tuple
        DopConeAngRef : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.PosRef = PosRef
        self.VelRef = VelRef
        self.DopConeAngRef = DopConeAngRef
        super(RMRefType, self).__init__(**kwargs)


class INCAType(Serializable):
    """Parameters for Imaging Near Closest Approach (INCA) image description."""
    _fields = (
        'TimeCAPoly', 'R_CA_SCP', 'FreqZero', 'DRateSFPoly', 'DopCentroidPoly', 'DopCentroidCOA')
    _required = ('TimeCAPoly', 'R_CA_SCP', 'FreqZero', 'DRateSFPoly')
    _numeric_format = {'R_CA_SCP': '0.17E', 'FreqZero': '0.17E'}
    # descriptors
    TimeCAPoly = SerializableDescriptor(
        'TimeCAPoly', Poly1DType, _required, strict=DEFAULT_STRICT,
        docstring='Polynomial function that yields *Time of Closest Approach* as function of '
                  'image column *(azimuth)* coordinate in meters. Time relative to '
                  'collection start in seconds.')  # type: Poly1DType
    R_CA_SCP = FloatDescriptor(
        'R_CA_SCP', _required, strict=DEFAULT_STRICT,
        docstring='*Range at Closest Approach (R_CA)* for the *Scene Center Point (SCP)* in meters.')  # type: float
    FreqZero = FloatDescriptor(
        'FreqZero', _required, strict=DEFAULT_STRICT,
        docstring=r'*RF frequency* :\math:`(f_0)` in Hz used for computing '
                  r'Doppler Centroid values. Typical :math:`f_0` '
                  r'set equal to center transmit frequency.')  # type: float
    DRateSFPoly = SerializableDescriptor(
        'DRateSFPoly', Poly2DType, _required, strict=DEFAULT_STRICT,
        docstring='Polynomial function that yields *Doppler Rate scale factor (DRSF)* '
                  'as a function of image location. Yields `DRSF` as a function of image '
                  'range coordinate ``(variable 1)`` and azimuth coordinate ``(variable 2)``. '
                  'Used to compute Doppler Rate at closest approach.')  # type: Poly2DType
    DopCentroidPoly = SerializableDescriptor(
        'DopCentroidPoly', Poly2DType, _required, strict=DEFAULT_STRICT,
        docstring='Polynomial function that yields Doppler Centroid value as a '
                  'function of image location *(fdop_DC)*. The *fdop_DC* is the '
                  'Doppler frequency at the peak signal response. The polynomial is a function '
                  'of image range coordinate ``(variable 1)`` and azimuth coordinate ``(variable 2)``. '
                  '*Note: Only used for Stripmap and Dynamic Stripmap collections.*')  # type: Poly2DType
    DopCentroidCOA = BooleanDescriptor(
        'DopCentroidCOA', _required, strict=DEFAULT_STRICT,
        docstring="""Flag indicating that the COA is at the peak signal :math`fdop_COA = fdop_DC`.
        
        * `True` - if Pixel COA at peak signal for all pixels.
        
        * `False` otherwise.
        
        *Note:* Only used for Stripmap and Dynamic Stripmap.""")  # type: bool

    def __init__(
            self,
            TimeCAPoly: Union[Poly1DType, numpy.ndarray, list, tuple] = None,
            R_CA_SCP: float = None,
            FreqZero: float = None,
            DRateSFPoly: Union[Poly2DType, numpy.ndarray, list, tuple] = None,
            DopCentroidPoly: Union[None, Poly2DType, numpy.ndarray, list, tuple] = None,
            DopCentroidCOA: bool = None,
            **kwargs):
        """

        Parameters
        ----------
        TimeCAPoly : Poly1DType|numpy.ndarray|list|tuple
        R_CA_SCP : float
        FreqZero : float
        DRateSFPoly : Poly2DType|numpy.ndarray|list|tuple
        DopCentroidPoly : Poly2DType|numpy.ndarray|list|tuple
        DopCentroidCOA : bool
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.TimeCAPoly = TimeCAPoly
        self.R_CA_SCP = R_CA_SCP
        self.FreqZero = FreqZero
        self.DRateSFPoly = DRateSFPoly
        self.DopCentroidPoly = DopCentroidPoly
        self.DopCentroidCOA = DopCentroidCOA
        super(INCAType, self).__init__(**kwargs)

    def _apply_reference_frequency(self, reference_frequency: float):
        if self.FreqZero is not None:
            self.FreqZero += reference_frequency


class RMAType(Serializable):
    """Parameters included when the image is formed using the Range Migration Algorithm."""
    _fields = ('RMAlgoType', 'ImageType', 'RMAT', 'RMCR', 'INCA')
    _required = ('RMAlgoType', 'ImageType')
    _choice = ({'required': True, 'collection': ('RMAT', 'RMCR', 'INCA')}, )
    # class variables
    _RM_ALGO_TYPE_VALUES = ('OMEGA_K', 'CSA', 'RG_DOP')
    # descriptors
    RMAlgoType = StringEnumDescriptor(
        'RMAlgoType', _RM_ALGO_TYPE_VALUES, _required, strict=DEFAULT_STRICT,
        docstring=r"""
        Identifies the type of migration algorithm used:

        * `OMEGA_K` - Algorithms that employ Stolt interpolation of the Kxt dimension. :math:`Kx = \sqrt{Kf^2 - Ky^2}`

        * `CSA` - Wave number algorithm that process two-dimensional chirp signals.

        * `RG_DOP` - Range-Doppler algorithms that employ *RCMC* in the compressed range domain.

        """)  # type: str
    RMAT = SerializableDescriptor(
        'RMAT', RMRefType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters for *RMA with Along Track (RMAT)* motion compensation.')  # type: RMRefType
    RMCR = SerializableDescriptor(
        'RMCR', RMRefType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters for *RMA with Cross Range (RMCR)* motion compensation.')  # type: RMRefType
    INCA = SerializableDescriptor(
        'INCA', INCAType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters for *Imaging Near Closest Approach (INCA)* image description.')  # type: INCAType

    def __init__(
            self,
            RMAlgoType: str = None,
            RMAT: Optional[RMRefType] = None,
            RMCR: Optional[RMRefType] = None,
            INCA: Optional[INCAType] = None,
            **kwargs):
        """

        Parameters
        ----------
        RMAlgoType : str
        RMAT : RMRefType
        RMCR : RMRefType
        INCA : INCAType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.RMAlgoType = RMAlgoType
        self.RMAT = RMAT
        self.RMCR = RMCR
        self.INCA = INCA
        super(RMAType, self).__init__(**kwargs)

    @property
    def ImageType(self) -> Optional[str]:
        """
        str: READ ONLY attribute. Identifies the specific RM image type / metadata type supplied. This is determined by
        returning the (first) attribute among `'RMAT', 'RMCR', 'INCA'` which is populated. `None` will be returned if
        none of them are populated.
        """

        for attribute in self._choice[0]['collection']:
            if getattr(self, attribute) is not None:
                return attribute
        return None

    def _derive_parameters(self, SCPCOA, Position, RadarCollection, ImageFormation):
        """
        Expected to be called from SICD parent.

        Parameters
        ----------
        SCPCOA : sarpy.io.complex.sicd_elements.SCPCOA.SCPCOAType
        Position : sarpy.io.complex.sicd_elements.Position.PositionType
        RadarCollection : sarpy.io.complex.sicd_elements.RadarCollection.RadarCollectionType
        ImageFormation : sarpy.io.complex.sicd_elements.ImageFormation.ImageFormationType

        Returns
        -------
        None
        """

        if SCPCOA is None:
            return

        scp = None if SCPCOA.ARPPos is None else SCPCOA.ARPPos.get_array()

        im_type = self.ImageType
        if im_type in ['RMAT', 'RMCR']:
            rm_ref = getattr(self, im_type)  # type: RMRefType
            if rm_ref.PosRef is None and SCPCOA.ARPPos is not None:
                rm_ref.PosRef = SCPCOA.ARPPos.copy()
            if rm_ref.VelRef is None and SCPCOA.ARPVel is not None:
                rm_ref.VelRef = SCPCOA.ARPVel.copy()
            if scp is not None and rm_ref.PosRef is not None and rm_ref.VelRef is not None:
                pos_ref = rm_ref.PosRef.get_array()
                vel_ref = rm_ref.VelRef.get_array()
                uvel_ref = vel_ref/norm(vel_ref)
                ulos = (scp - pos_ref)  # it absolutely could be that scp = pos_ref
                ulos_norm = norm(ulos)
                if ulos_norm > 0:
                    ulos /= ulos_norm
                    if rm_ref.DopConeAngRef is None:
                        rm_ref.DopConeAngRef = numpy.rad2deg(numpy.arccos(numpy.dot(uvel_ref, ulos)))
        elif im_type == 'INCA':
            if scp is not None and self.INCA.TimeCAPoly is not None and \
                    Position is not None and Position.ARPPoly is not None:
                t_zero = self.INCA.TimeCAPoly.Coefs[0]
                ca_pos = Position.ARPPoly(t_zero)
                if self.INCA.R_CA_SCP is None:
                    self.INCA.R_CA_SCP = norm(ca_pos - scp)
            if self.INCA.FreqZero is None:
                self.INCA.FreqZero = _get_center_frequency(RadarCollection, ImageFormation)

    def _apply_reference_frequency(self, reference_frequency: float):
        """
        If the reference frequency is used, adjust the necessary fields accordingly.
        Expected to be called by SICD parent.

        Parameters
        ----------
        reference_frequency : float
            The reference frequency.

        Returns
        -------
        None
        """

        if self.INCA is not None:
            # noinspection PyProtectedMember
            self.INCA._apply_reference_frequency(reference_frequency)




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class SITE(TREElement):
    def __init__(self, value):
        super(SITE, self).__init__()
        self.add_field('SCLCTN_NAME', 's', 20, value)
        self.add_field('SDESCRIPTION', 's', 255, value)
        self.add_field('SITE_NUM', 's', 3, value)
        self.add_field('SCN_NUM', 's', 3, value)
        self.add_field('SCLCTN_STDATE', 's', 8, value)
        self.add_field('SCLCTN_SPDATE', 's', 8, value)
        self.add_field('SCN_CNTR', 's', 11, value)
        self.add_field('ALTITUDE', 's', 5, value)
        self.add_field('SCN_CONTENT', 's', 50, value)
        self.add_field('BGRND_TYPE', 's', 50, value)
        self.add_field('SITE_COV', 's', 1, value)


class CLCTNBType(TREElement):
    def __init__(self, value):
        super(CLCTNBType, self).__init__()
        self.add_field('VERNUM', 's', 4, value)
        self.add_field('CLCTN_NAME', 's', 25, value)
        self.add_field('CLCTN_DESCRIPT', 's', 255, value)
        self.add_field('CLCTN_STDATE', 's', 8, value)
        self.add_field('CLCTN_SPDATE', 's', 8, value)
        self.add_field('CLCTN_LOC', 's', 11, value)
        self.add_field('SITE', 's', 15, value)
        self.add_field('COUNTRY', 's', 2, value)
        self.add_field('SPONSOR', 's', 20, value)
        self.add_field('PERSONNEL', 's', 100, value)
        self.add_field('NUM_SITES', 'd', 1, value)
        self.add_loop('SITEs', self.NUM_SITES, SITE, value)
        self.add_field('SCLCTN_Z_OFF', 's', 3, value)
        self.add_field('SECURITY', 's', 7, value)
        self.add_field('SCG', 's', 15, value)
        self.add_field('FLIGHT_NUM', 's', 2, value)
        self.add_field('PASS_NUM', 's', 2, value)
        self.add_field('WX_STATION', 's', 20, value)
        self.add_field('WX_OVERVIEW', 's', 15, value)
        self.add_field('WX_FILE', 's', 30, value)


class CLCTNB(TREExtension):
    _tag_value = 'CLCTNB'
    _data_type = CLCTNBType



"""
Unified methods of projection between sicd pixel coordinates,
some ortho-rectified pixel grid coordinates, and geophysical coordinates
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

import abc
import logging

import numpy
from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.geometry.geocoords import geodetic_to_ecf, ecf_to_geodetic, wgs_84_norm
from sarpy.processing.rational_polynomial import SarpyRatPolyError, \
    get_rational_poly_2d, get_rational_poly_3d, CombinedRationalPolynomial


logger = logging.getLogger(__name__)

_PIXEL_METHODOLOGY = ('MAX', 'MIN', 'MEAN', 'GEOM_MEAN')


class ProjectionHelper(abc.ABC):
    """
    Abstract helper class which defines the projection interface for
    ortho-rectification usage for a sicd type object.
    """

    __slots__ = ('_sicd', '_row_spacing', '_col_spacing', '_default_pixel_method')

    def __init__(self, sicd, row_spacing=None, col_spacing=None, default_pixel_method='GEOM_MEAN'):
        r"""

        Parameters
        ----------
        sicd : SICDType
            The sicd object
        row_spacing : None|float
            The row pixel spacing. If not provided, this will default according
            to `default_pixel_method`.
        col_spacing : None|float
            The row pixel spacing. If not provided, this will default according
            to `default_pixel_method`.
        default_pixel_method : str
            Must be one of ('MAX', 'MIN', 'MEAN', 'GEOM_MEAN'). This determines
            the default behavior for row_spacing/col_spacing. The default value for
            row/column spacing will be the implied function applied to the range
            and azimuth ground resolution. Note that geometric mean is defined as
            :math:`\sqrt(x*x + y*y)`
        """

        self._row_spacing = None
        self._col_spacing = None
        default_pixel_method = default_pixel_method.upper()
        if default_pixel_method not in _PIXEL_METHODOLOGY:
            raise ValueError(
                'default_pixel_method got invalid value {}. Must be one '
                'of {}'.format(default_pixel_method, _PIXEL_METHODOLOGY))
        self._default_pixel_method = default_pixel_method

        if not isinstance(sicd, SICDType):
            raise TypeError('sicd must be a SICDType instance. Got type {}'.format(type(sicd)))
        if not sicd.can_project_coordinates():
            raise ValueError('Ortho-rectification requires the SICD ability to project coordinates.')
        sicd.define_coa_projection(override=False)
        self._sicd = sicd
        self.row_spacing = row_spacing
        self.col_spacing = col_spacing

    @property
    def sicd(self):
        """
        SICDType: The sicd structure.
        """

        return self._sicd

    @property
    def row_spacing(self):
        """
        float: The row pixel spacing
        """

        return self._row_spacing

    @row_spacing.setter
    def row_spacing(self, value):
        """
        Set the row pixel spacing value. Setting to None will result in a
        default value derived from the SICD structure being used.

        Parameters
        ----------
        value : None|float

        Returns
        -------
        None
        """

        if value is None:
            if (self.sicd.RadarCollection.Area is None or self.sicd.RadarCollection.Area.Plane is None):
                self._row_spacing = self._get_sicd_ground_pixel()
            else:
                self._row_spacing = self.sicd.RadarCollection.Area.Plane.XDir.LineSpacing
        else:
            value = float(value)
            if value <= 0:
                raise ValueError('row pixel spacing must be positive.')
            self._row_spacing = float(value)

    @property
    def col_spacing(self):
        """
        float: The column pixel spacing
        """

        return self._col_spacing

    @col_spacing.setter
    def col_spacing(self, value):
        """
        Set the col pixel spacing value. Setting to NOne will result in a
        default value derived from the SICD structure being used.

        Parameters
        ----------
        value : None|float

        Returns
        -------
        None
        """

        if value is None:
            if (self.sicd.RadarCollection.Area is None or self.sicd.RadarCollection.Area.Plane is None):
                self._col_spacing = self._get_sicd_ground_pixel()
            else:
                self._col_spacing = self.sicd.RadarCollection.Area.Plane.YDir.SampleSpacing
        else:
            value = float(value)
            if value <= 0:
                raise ValueError('column pixel spacing must be positive.')
            self._col_spacing = float(value)

    def _get_sicd_ground_pixel(self):
        """
        Gets the SICD ground pixel size.

        Returns
        -------
        float
        """

        ground_row_ss, ground_col_ss = self.sicd.get_ground_resolution()
        if self._default_pixel_method == 'MIN':
            return min(ground_row_ss, ground_col_ss)
        elif self._default_pixel_method == 'MAX':
            return max(ground_row_ss, ground_col_ss)
        elif self._default_pixel_method == 'MEAN':
            return 0.5*(ground_row_ss + ground_col_ss)
        elif self._default_pixel_method == 'GEOM_MEAN':
            return float(numpy.sqrt(ground_row_ss*ground_row_ss + ground_col_ss*ground_col_ss))
        else:
            raise ValueError('Got unhandled default_pixel_method {}'.format(self._default_pixel_method))

    @staticmethod
    def _reshape(array, final_dimension):
        """
        Reshape the input so that the output is two-dimensional with final
        dimension given by `final_dimension`.

        Parameters
        ----------
        array : numpy.ndarray|list|tuple
        final_dimension : int

        Returns
        -------
        (numpy.ndarray, tuple)
            The reshaped data array and original shape.
        """

        if not isinstance(array, numpy.ndarray):
            array = numpy.array(array, dtype=numpy.float64)
        if array.ndim < 1 or array.shape[-1] != final_dimension:
            raise ValueError(
                'ortho_coords must be at least one dimensional with final dimension '
                'of size {}.'.format(final_dimension))

        o_shape = array.shape

        if array.ndim != 2:
            array = numpy.reshape(array, (-1, final_dimension))
        return array, o_shape

    @abc.abstractmethod
    def ecf_to_ortho(self, coords):
        """
        Gets the `(ortho_row, ortho_column)` coordinates in the ortho-rectified
        system for the provided physical coordinates in ECF `(X, Y, Z)` coordinates.

        Parameters
        ----------
        coords : numpy.ndarray|list|tuple

        Returns
        -------
        numpy.ndarray
        """

        raise NotImplementedError

    @abc.abstractmethod
    def ecf_to_pixel(self, coords):
        """
        Gets the `(pixel_row, pixel_column)` coordinates for the provided physical
        coordinates in ECF `(X, Y, Z)` coordinates.

        Parameters
        ----------
        coords : numpy.ndarray|list|tuple

        Returns
        -------
        numpy.ndarray
        """

        raise NotImplementedError

    @abc.abstractmethod
    def ll_to_ortho(self, ll_coords):
        """
        Gets the `(ortho_row, ortho_column)` coordinates in the ortho-rectified
        system for the provided physical coordinates in `(Lat, Lon)` coordinates.

        Note that there is inherent ambiguity when handling the missing elevation,
        and the effect is likely methodology dependent.

        Parameters
        ----------
        ll_coords : numpy.ndarray|list|tuple

        Returns
        -------
        numpy.ndarray
        """

        raise NotImplementedError

    @abc.abstractmethod
    def llh_to_ortho(self, llh_coords):
        """
        Gets the `(ortho_row, ortho_column)` coordinates in the ortho-rectified
        system for the provided physical coordinates in `(Lat, Lon, HAE)`
        coordinates.

        Parameters
        ----------
        llh_coords : numpy.ndarray|list|tuple

        Returns
        -------
        numpy.ndarray
        """

        raise NotImplementedError

    @abc.abstractmethod
    def pixel_to_ortho(self, pixel_coords):
        """
        Gets the ortho-rectified indices for the point(s) in pixel coordinates.

        Parameters
        ----------
        pixel_coords : numpy.ndarray|list|tuple

        Returns
        -------
        numpy.ndarray
        """

        raise NotImplementedError

    @abc.abstractmethod
    def pixel_to_ecf(self, pixel_coords):
        """
        Gets the ECF coordinates for the point(s) in pixel coordinates.

        Parameters
        ----------
        pixel_coords : numpy.ndarray|list|tuple

        Returns
        -------
        numpy.ndarray
        """

        raise NotImplementedError

    @abc.abstractmethod
    def ortho_to_ecf(self, ortho_coords):
        """
        Get the ecf coordinates for the point(s) in ortho-rectified coordinates.

        Parameters
        ----------
        ortho_coords : numpy.ndarray
            Point(s) in the ortho-rectified coordinate system, of the form
            `(ortho_row, ortho_column)`.

        Returns
        -------
        numpy.ndarray
        """

        raise NotImplementedError

    def ortho_to_llh(self, ortho_coords):
        """
        Get the lat/lon/hae coordinates for the point(s) in ortho-rectified coordinates.

        Parameters
        ----------
        ortho_coords : numpy.ndarray
            Point(s) in the ortho-rectified coordinate system, of the form
            `(ortho_row, ortho_column)`.

        Returns
        -------
        numpy.ndarray
        """

        ecf = self.ortho_to_ecf(ortho_coords)
        return ecf_to_geodetic(ecf)

    @abc.abstractmethod
    def ortho_to_pixel(self, ortho_coords):
        """
        Get the pixel indices for the point(s) in ortho-rectified coordinates.

        Parameters
        ----------
        ortho_coords : numpy.ndarray
            Point(s) in the ortho-rectified coordinate system, of the form
            `(ortho_row, ortho_column)`.

        Returns
        -------
        numpy.ndarray
            The array of indices, of the same shape as `new_coords`, which indicate
            `(row, column)` pixel (fractional) indices.
        """

        raise NotImplementedError

    def get_pixel_array_bounds(self, coords):
        """
        Extract integer bounds of the input array, expected to have final dimension
        of size 2.

        Parameters
        ----------
        coords : numpy.ndarray

        Returns
        -------
        numpy.ndarray
            Of the form `(min_row, max_row, min_column, max_column)`.
        """

        coords, o_shape = self._reshape(coords, 2)
        return numpy.array(
            (numpy.ceil(numpy.nanmin(coords[:, 0], axis=0)),
             numpy.floor(numpy.nanmax(coords[:, 0], axis=0)),
             numpy.ceil(numpy.nanmin(coords[:, 1], axis=0)),
             numpy.floor(numpy.nanmax(coords[:, 1], axis=0))), dtype=numpy.int64)


class PGProjection(ProjectionHelper):
    """
    Class which helps perform the Planar Grid (i.e. Ground Plane) ortho-rectification
    for a sicd-type object using the SICD projection model directly.
    """

    __slots__ = (
        '_reference_point', '_reference_pixels', '_row_vector', '_col_vector', '_normal_vector', '_reference_hae')

    def __init__(self, sicd, reference_point=None, reference_pixels=None, normal_vector=None, row_vector=None,
                 col_vector=None, row_spacing=None, col_spacing=None,
                 default_pixel_method='GEOM_MEAN'):
        r"""

        Parameters
        ----------
        sicd : SICDType
            The sicd object
        reference_point : None|numpy.ndarray
            The reference point (origin) of the planar grid. If None, a default
            derived from the SICD will be used.
        reference_pixels : None|numpy.ndarray
            The projected pixel
        normal_vector : None|numpy.ndarray
            The unit normal vector of the plane.
        row_vector : None|numpy.ndarray
            The vector defining increasing column direction. If None, a default
            derived from the SICD will be used.
        col_vector : None|numpy.ndarray
            The vector defining increasing column direction. If None, a default
            derived from the SICD will be used.
        row_spacing : None|float
            The row pixel spacing.
        col_spacing : None|float
            The column pixel spacing.
        default_pixel_method : str
            Must be one of ('MAX', 'MIN', 'MEAN', 'GEOM_MEAN'). This determines
            the default behavior for row_spacing/col_spacing. The default value for
            row/column spacing will be the implied function applied to the range
            and azimuth ground resolution. Note that geometric mean is defined as
            :math:`\sqrt(x*x + y*y)`
        """

        self._reference_point = None
        self._reference_hae = None
        self._reference_pixels = None
        self._normal_vector = None
        self._row_vector = None
        self._col_vector = None
        ProjectionHelper.__init__(
            self, sicd, row_spacing=row_spacing, col_spacing=col_spacing, default_pixel_method=default_pixel_method)
        self.set_reference_point(reference_point=reference_point)
        self.set_reference_pixels(reference_pixels=reference_pixels)
        self.set_plane_frame(
            normal_vector=normal_vector, row_vector=row_vector, col_vector=col_vector)

    @property
    def reference_point(self):
        # type: () -> numpy.ndarray
        """
        numpy.ndarray: The grid reference point.
        """

        return self._reference_point

    @property
    def reference_pixels(self):
        # type: () -> numpy.ndarray
        """
        numpy.ndarray: The ortho-rectified pixel coordinates of the grid reference point.
        """

        return self._reference_pixels

    @property
    def normal_vector(self):
        # type: () -> numpy.ndarray
        """
        numpy.ndarray: The normal vector.
        """

        return self._normal_vector

    def set_reference_point(self, reference_point=None):
        """
        Sets the reference point, which must be provided in ECF coordinates.

        Parameters
        ----------
        reference_point : None|numpy.ndarray
            The reference point (origin) of the planar grid. If None, then the
            `sicd.GeoData.SCP.ECF` will be used.

        Returns
        -------
        None
        """

        if reference_point is None:
            if (self.sicd.RadarCollection.Area is None or self.sicd.RadarCollection.Area.Plane is None):
                reference_point = self.sicd.GeoData.SCP.ECF.get_array()
            else:
                reference_point = self.sicd.RadarCollection.Area.Plane.RefPt.ECF.get_array()

        if not (isinstance(reference_point, numpy.ndarray) and reference_point.ndim == 1
                and reference_point.size == 3):
            raise ValueError('reference_point must be a vector of size 3.')
        self._reference_point = reference_point
        # set the reference hae
        ref_llh = ecf_to_geodetic(reference_point)
        self._reference_hae = ref_llh[2]

    def set_reference_pixels(self, reference_pixels=None):
        """
        Sets the reference point, which must be provided in ECF coordinates.

        Parameters
        ----------
        reference_pixels : None|numpy.ndarray
            The ortho-rectified pixel coordinates for the reference point (origin) of the planar grid.
            If None, then the (0, 0) will be used.

        Returns
        -------
        None
        """

        if reference_pixels is None:
            if (self.sicd.RadarCollection.Area is None or self.sicd.RadarCollection.Area.Plane is None):
                reference_pixels = numpy.zeros((2, ), dtype='float64')
            else:
                reference_pixels = numpy.array([
                    self.sicd.RadarCollection.Area.Plane.RefPt.Line,
                    self.sicd.RadarCollection.Area.Plane.RefPt.Sample],
                    dtype='float64')

        if not (isinstance(reference_pixels, numpy.ndarray) and reference_pixels.ndim == 1
                and reference_pixels.size == 2):
            raise ValueError('reference_pixels must be a vector of size 2.')
        self._reference_pixels = reference_pixels

    @property
    def row_vector(self):
        """
        numpy.ndarray: The grid increasing row direction (ECF) unit vector.
        """

        return self._row_vector

    @property
    def col_vector(self):
        """
        numpy.ndarray: The grid increasing column direction (ECF) unit vector.
        """

        return self._col_vector

    @property
    def reference_hae(self):
        """
        float: The height above the ellipsoid of the reference point.
        """

        return self._reference_hae

    def set_plane_frame(self, normal_vector=None, row_vector=None, col_vector=None):
        """
        Set the plane unit normal, and the row and column vectors, in ECF coordinates.
        Note that the perpendicular component of col_vector with respect to the
        row_vector will be used.

        If `normal_vector`, `row_vector`, and `col_vector` are all `None`, then
        the normal to the Earth tangent plane at the reference point is used for
        `normal_vector`. The `row_vector` will be defined as the perpendicular
        component of `sicd.Grid.Row.UVectECF` to `normal_vector`. The `column_vector`
        will be defined as the component of `sicd.Grid.Col.UVectECF` perpendicular
        to both `normal_vector` and `row_vector`.

        If only `normal_vector` is supplied, then the `row_vector` and `column_vector`
        will be defined similarly as the perpendicular components of
        `sicd.Grid.Row.UVectECF` and `sicd.Grid.Col.UVectECF`.

        Otherwise, all vectors supplied will be normalized, but are required to be
        mutually perpendicular. If only two vectors are supplied, then the third
        will be determined.

        Parameters
        ----------
        normal_vector : None|numpy.ndarray
            The vector defining the outward unit normal in ECF coordinates.
        row_vector : None|numpy.ndarray
            The vector defining increasing column direction.
        col_vector : None|numpy.ndarray
            The vector defining increasing column direction.

        Returns
        -------
        None
        """

        def normalize(vec, name, perp=None):
            if not isinstance(vec, numpy.ndarray):
                vec = numpy.array(vec, dtype=numpy.float64)
            if not (isinstance(vec, numpy.ndarray) and vec.ndim == 1 and vec.size == 3):
                raise ValueError('{} vector must be a numpy.ndarray of dimension 1 and size 3.'.format(name))
            vec = numpy.copy(vec)
            if perp is None:
                pass
            elif isinstance(perp, numpy.ndarray):
                vec = vec - perp*(perp.dot(vec))
            else:
                for entry in perp:
                    vec = vec - entry*(entry.dot(vec))

            norm = numpy.linalg.norm(vec)
            if norm == 0:
                raise ValueError('{} vector cannot be the zero vector.'.format(name))
            elif norm != 1:
                vec = vec/norm  # avoid modifying row_vector def exterior to this class
            return vec

        def check_perp(vec1, vec2, name1, name2, tolerance=1e-6):
            if abs(vec1.dot(vec2)) > tolerance:
                raise ValueError('{} vector and {} vector are not perpendicular'.format(name1, name2))

        if self._reference_point is None:
            raise ValueError('This requires that reference point is previously set.')

        if normal_vector is None and row_vector is None and col_vector is None:
            if (self.sicd.RadarCollection.Area is None or self.sicd.RadarCollection.Area.Plane is None):
                self._normal_vector = wgs_84_norm(self.reference_point)
                self._row_vector = normalize(
                    self.sicd.Grid.Row.UVectECF.get_array(), 'row', perp=self.normal_vector)
                self._col_vector = normalize(
                    self.sicd.Grid.Col.UVectECF.get_array(), 'column', perp=(self.normal_vector, self.row_vector))
            else:
                self._row_vector = self.sicd.RadarCollection.Area.Plane.XDir.UVectECF.get_array()
                self._col_vector = normalize(
                    self.sicd.RadarCollection.Area.Plane.YDir.UVectECF.get_array(), 'col', perp=self._row_vector)
                self._normal_vector = numpy.cross(self._row_vector, self._col_vector)
        elif normal_vector is not None and row_vector is None and col_vector is None:
            self._normal_vector = normalize(normal_vector, 'normal')
            self._row_vector = normalize(
                self.sicd.Grid.Row.UVectECF.get_array(), 'row', perp=self.normal_vector)
            self._col_vector = normalize(
                self.sicd.Grid.Col.UVectECF.get_array(), 'column', perp=(self.normal_vector, self.row_vector))
        elif normal_vector is None:
            if row_vector is None or col_vector is None:
                raise ValueError('normal_vector is not defined, so both row_vector and col_vector must be.')
            row_vector = normalize(row_vector, 'row')
            col_vector = normalize(col_vector, 'col')
            check_perp(row_vector, col_vector, 'row', 'col')
            self._row_vector = row_vector
            self._col_vector = col_vector
            self._normal_vector = numpy.cross(row_vector, col_vector)
        elif col_vector is None:
            if row_vector is None:
                raise ValueError('col_vector is not defined, so both normal_vector and row_vector must be.')
            normal_vector = normalize(normal_vector, 'normal')
            row_vector = normalize(row_vector, 'row')
            check_perp(normal_vector, row_vector, 'normal', 'row')
            self._normal_vector = normal_vector
            self._row_vector = row_vector
            self._col_vector = numpy.cross(self.normal_vector, self.row_vector)
        elif row_vector is None:
            normal_vector = normalize(normal_vector, 'normal')
            col_vector = normalize(col_vector, 'col')
            check_perp(normal_vector, col_vector, 'normal', 'col')
            self._normal_vector = normal_vector
            self._col_vector = col_vector
            self._row_vector = numpy.cross(self.col_vector, self.normal_vector)
        else:
            normal_vector = normalize(normal_vector, 'normal')
            row_vector = normalize(row_vector, 'row')
            col_vector = normalize(col_vector, 'col')
            check_perp(normal_vector, row_vector, 'normal', 'row')
            check_perp(normal_vector, col_vector, 'normal', 'col')
            check_perp(row_vector, col_vector, 'row', 'col')
            self._normal_vector = normal_vector
            self._row_vector = row_vector
            self._col_vector = col_vector
        # check for outward unit norm
        if numpy.dot(self.normal_vector, self.reference_point) < 0:
            logger.warning(
                'The normal vector appears to be outward pointing, so reversing.')
            self._normal_vector *= -1

    def plane_ecf_to_ortho(self, coords):
        """
        Converts ECF coordinates **known to be in the ground plane** to ortho grid coordinates.

        Parameters
        ----------
        coords : numpy.ndarray

        Returns
        -------
        numpy.ndarray
        """

        coords, o_shape = self._reshape(coords, 3)
        diff = coords - self.reference_point
        if len(o_shape) == 1:
            out = numpy.zeros((2, ), dtype=numpy.float64)
            out[0] = self._reference_pixels[0] + numpy.sum(diff*self.row_vector)/self.row_spacing
            out[1] = self._reference_pixels[1] + numpy.sum(diff*self.col_vector)/self.col_spacing
        else:
            out = numpy.zeros((coords.shape[0], 2), dtype=numpy.float64)
            out[:, 0] = self._reference_pixels[0] + numpy.sum(diff*self.row_vector, axis=1)/self.row_spacing
            out[:, 1] = self._reference_pixels[1] + numpy.sum(diff*self.col_vector, axis=1)/self.col_spacing
            out = numpy.reshape(out, o_shape[:-1] + (2, ))
        return out

    def ecf_to_ortho(self, coords):
        return self.pixel_to_ortho(self.ecf_to_pixel(coords))

    def ecf_to_pixel(self, coords):
        pixel, _, _ = self.sicd.project_ground_to_image(coords, tolerance=1e-6, max_iterations=40)
        return pixel

    def ll_to_ortho(self, ll_coords):
        """
        Gets the `(ortho_row, ortho_column)` coordinates in the ortho-rectified
        system for the provided physical coordinates in `(Lat, Lon)` coordinates.
        In this case, the missing altitude will be set to `reference_hae`, which
        is imperfect.

        Parameters
        ----------
        ll_coords : numpy.ndarray|list|tuple

        Returns
        -------
        numpy.ndarray
        """

        ll_coords, o_shape = self._reshape(ll_coords, 2)
        llh_temp = numpy.zeros((ll_coords.shape[0], 3), dtype=numpy.float64)
        llh_temp[:, :2] = ll_coords
        llh_temp[:, 2] = self.reference_hae
        llh_temp = numpy.reshape(llh_temp, o_shape[:-1] + (3, ))
        return self.llh_to_ortho(llh_temp)

    def llh_to_ortho(self, llh_coords):
        llh_coords, o_shape = self._reshape(llh_coords, 3)
        ground = geodetic_to_ecf(llh_coords)
        return self.ecf_to_ortho(numpy.reshape(ground, o_shape))

    def ortho_to_ecf(self, ortho_coords):
        ortho_coords, o_shape = self._reshape(ortho_coords, 2)
        xs = (ortho_coords[:, 0] - self._reference_pixels[0])*self.row_spacing
        ys = (ortho_coords[:, 1] - self._reference_pixels[1])*self.col_spacing
        if xs.ndim == 0:
            coords = self.reference_point + xs*self.row_vector + ys*self.col_vector
        else:
            coords = self.reference_point + numpy.outer(xs, self.row_vector) + \
                     numpy.outer(ys, self.col_vector)
        return numpy.reshape(coords, o_shape[:-1] + (3, ))

    def ortho_to_pixel(self, ortho_coords):
        ortho_coords, o_shape = self._reshape(ortho_coords, 2)
        pixel, _, _ = self.sicd.project_ground_to_image(self.ortho_to_ecf(ortho_coords), tolerance=1e-3,
                                                        max_iterations=25)
        return numpy.reshape(pixel, o_shape)

    def pixel_to_ortho(self, pixel_coords):
        return self.plane_ecf_to_ortho(self.pixel_to_ecf(pixel_coords))

    def pixel_to_ecf(self, pixel_coords):
        return self.sicd.project_image_to_ground(
            pixel_coords, projection_type='PLANE',
            gref=self.reference_point, ugpn=self.normal_vector)


class PGRatPolyProjection(PGProjection):

    __slots__ = (
        '_reference_point', '_reference_pixels', '_row_vector', '_col_vector',
        '_normal_vector', '_reference_hae',
        '_row_samples', '_col_samples', '_alt_samples', '_alt_span',
        '_ecf_to_pixel_func', '_pixel_to_ortho_func', '_ortho_to_pixel_func')

    def __init__(self, sicd, reference_point=None, reference_pixels=None, normal_vector=None, row_vector=None,
                 col_vector=None, row_spacing=None, col_spacing=None,
                 default_pixel_method='GEOM_MEAN',
                 row_samples=51, col_samples=51, alt_samples=11, alt_span=250):
        r"""

        Parameters
        ----------
        sicd : SICDType
            The sicd object
        reference_point : None|numpy.ndarray
            The reference point (origin) of the planar grid. If None, a default
            derived from the SICD will be used.
        reference_pixels : None|numpy.ndarray
            The projected pixel
        normal_vector : None|numpy.ndarray
            The unit normal vector of the plane.
        row_vector : None|numpy.ndarray
            The vector defining increasing column direction. If None, a default
            derived from the SICD will be used.
        col_vector : None|numpy.ndarray
            The vector defining increasing column direction. If None, a default
            derived from the SICD will be used.
        row_spacing : None|float
            The row pixel spacing.
        col_spacing : None|float
            The column pixel spacing.
        default_pixel_method : str
            Must be one of ('MAX', 'MIN', 'MEAN', 'GEOM_MEAN'). This determines
            the default behavior for row_spacing/col_spacing. The default value for
            row/column spacing will be the implied function applied to the range
            and azimuth ground resolution. Note that geometric mean is defined as
            :math:`\sqrt(x*x + y*y)`
        row_samples : int
            How many row samples to use in fitting
        col_samples : int
            How many column samples to use in fitting
        alt_samples : int
            How many altitude samples to use in fitting
        alt_span : int|float
            Fitting for reference point hae +/- alt_span information.
        """

        self._ecf_to_pixel_func = None
        self._pixel_to_ortho_func = None
        self._ortho_to_pixel_func = None
        self._row_samples = int(row_samples)
        self._col_samples = int(col_samples)
        self._alt_samples = int(alt_samples)
        self._alt_span = float(alt_span)

        PGProjection.__init__(
            self, sicd, reference_point=reference_point, reference_pixels=reference_pixels,
            normal_vector=normal_vector, row_vector=row_vector, col_vector=col_vector,
            row_spacing=row_spacing, col_spacing=col_spacing, default_pixel_method=default_pixel_method)
        self.perform_rational_poly_fitting()

    def _perform_ecf_func_fitting(self):
        num_rows = self.sicd.ImageData.NumRows
        num_cols = self.sicd.ImageData.NumCols

        row_array = numpy.linspace(0, num_rows-1, self._row_samples)
        col_array = numpy.linspace(0, num_cols-1, self._col_samples)
        hae_array = self.reference_hae + numpy.linspace(
            -self._alt_span, self._alt_span, self._alt_samples)

        row_col_grid = numpy.empty((row_array.size, col_array.size, 2), dtype='float64')
        row_col_grid[:, :, 1], row_col_grid[:, :, 0] = numpy.meshgrid(col_array, row_array)

        ECF_data = numpy.empty((row_array.size, col_array.size, hae_array.size, 3))
        for i, hae0 in enumerate(hae_array):
            ECF_data[:, :, i, :] = self.sicd.project_image_to_ground(row_col_grid, projection_type='HAE', hae0=hae0)

        if not numpy.all(numpy.isfinite(ECF_data)):
            raise SarpyRatPolyError(
                'NaN values are encountered when projecting across the image area,\n\t'
                'this SICD is not a good candidate for projection using rational polynomials')
        row_func = get_rational_poly_3d(
            ECF_data[:, :, :, 0].flatten(), ECF_data[:, :, :, 1].flatten(), ECF_data[:, :, :, 2].flatten(),
            numpy.stack([row_col_grid[:, :, 0] for _ in range(self._alt_samples)], axis=2).flatten(), order=3)
        col_func = get_rational_poly_3d(
            ECF_data[:, :, :, 0].flatten(), ECF_data[:, :, :, 1].flatten(), ECF_data[:, :, :, 2].flatten(),
            numpy.stack([row_col_grid[:, :, 1] for _ in range(self._alt_samples)], axis=2).flatten(), order=3)
        self._ecf_to_pixel_func = CombinedRationalPolynomial(row_func, col_func)

    def _perform_pixel_fitting(self):
        num_rows = self.sicd.ImageData.NumRows
        num_cols = self.sicd.ImageData.NumCols
        row_array = numpy.linspace(0, num_rows-1, 2*self._row_samples)
        col_array = numpy.linspace(0, num_cols-1, 2*self._col_samples)
        pixel_data = numpy.empty((row_array.size, col_array.size, 2), dtype='float64')
        pixel_data[:, :, 1], pixel_data[:, :, 0] = numpy.meshgrid(col_array, row_array)

        ortho_data = PGProjection.pixel_to_ortho(self, pixel_data)
        pix_to_orth_row = get_rational_poly_2d(
            pixel_data[:, :, 0].flatten(), pixel_data[:, :, 1].flatten(),
            ortho_data[:, :, 0], order=3)
        pix_to_orth_col = get_rational_poly_2d(
            pixel_data[:, :, 0].flatten(), pixel_data[:, :, 1].flatten(),
            ortho_data[:, :, 1], order=3)
        self._pixel_to_ortho_func = CombinedRationalPolynomial(pix_to_orth_row, pix_to_orth_col)

        orth_to_pix_row = get_rational_poly_2d(
            ortho_data[:, :, 0].flatten(), ortho_data[:, :, 1].flatten(),
            pixel_data[:, :, 0], order=3)
        orth_to_pix_col = get_rational_poly_2d(
            ortho_data[:, :, 0].flatten(), ortho_data[:, :, 1].flatten(),
            pixel_data[:, :, 1], order=3)
        self._ortho_to_pixel_func = CombinedRationalPolynomial(orth_to_pix_row, orth_to_pix_col)

    def perform_rational_poly_fitting(self):
        """
        Defined the rational polynomial functions via fitting.
        """

        self._perform_ecf_func_fitting()
        self._perform_pixel_fitting()

    def ecf_to_ortho(self, coords):
        return self.pixel_to_ortho(self.ecf_to_pixel(coords))

    def ecf_to_pixel(self, coords):
        return self._ecf_to_pixel_func(coords, combine=True)

    def ortho_to_pixel(self, ortho_coords):
        return self._ortho_to_pixel_func(ortho_coords, combine=True)

    def pixel_to_ortho(self, pixel_coords):
        return self._pixel_to_ortho_func(pixel_coords, combine=True)



"""
Types implementing the SIDD 3.0 Measurement Parameters
"""
__classification__ = 'UNCLASSIFIED'

# SIDD 3.0 reuses the SIDD 2.0 types.  Make those symbols available in this module.
from sarpy.io.product.sidd2_elements.Measurement import (
    BaseProjectionType,
    MeasurableProjectionType,
    ProductPlaneType,
    PlaneProjectionType,
    GeographicProjectionType,
    CylindricalProjectionType,
    PolynomialProjectionType,
    MeasurementType,
)

__REUSED__ = (  # to avoid unused import lint errors
    BaseProjectionType,
    MeasurableProjectionType,
    ProductPlaneType,
    PlaneProjectionType,
    GeographicProjectionType,
    CylindricalProjectionType,
    PolynomialProjectionType,
    MeasurementType,
)



"""
General purpose rational polynomial tools
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

import logging
from typing import List, Tuple, Sequence, Optional, Union

import numpy
from numpy.polynomial import polynomial
from scipy.linalg import lstsq, LinAlgError

from sarpy.compliance import SarpyError

logger = logging.getLogger(__name__)


class SarpyRatPolyError(SarpyError):
    """A custom exception class for rational polynomial fitting errors."""


#################
# helper functions

def _get_num_variables(coeff_list: Sequence[Union[int, Tuple[int, ...]]]) -> int:
    """
    Determine the number of variables by inspection of the coefficient list

    Parameters
    ----------
    coeff_list : Sequence

    Returns
    -------
    int
    """

    variables = None
    for entry in coeff_list:
        if isinstance(entry, int):
            if variables is None:
                variables = 1
            elif variables != 1:
                raise ValueError('Entry order mismatch')
        else:
            if variables is None:
                variables = len(entry)
            elif variables != len(entry):
                raise ValueError('Entry order mismatch')
    if variables is None:
        raise ValueError('Unable to determine the number of variables')
    return variables


def _map_list_to_poly_matrix(coeffs: Sequence[float], coeff_list: Sequence[Tuple[int, ...]]) -> numpy.ndarray:
    """
    Maps the coefficients and coefficient listing to corresponding
    numpy polynomial coefficient matrix.

    Parameters
    ----------
    coeffs : Sequence[float]
    coeff_list : Sequence[Tuple[int, ...]]

    Returns
    -------
    coefficient_array : numpy.ndarray
    """

    variables = _get_num_variables(coeff_list)

    matrix_shape = []
    for i in range(variables):
        matrix_shape.append(max(entry[i] for entry in coeff_list)+1)

    coefficient_array = numpy.zeros(tuple(matrix_shape), dtype='float64')
    for i, entry in enumerate(coeff_list):
        coefficient_array[entry] = coeffs[i]
    return coefficient_array


def get_default_coefficient_ordering(variables: int, order: int) -> Sequence[Tuple[int, ...]]:
    """
    Gets a sensible coefficient ordering of a polynomial of given number of
    variables and order.

    Parameters
    ----------
    variables : int
    order : int

    Returns
    -------
    coefficient_list : Tuple[Tuple[int, ...]]
        List of the form `[(exponent 0, exponent 1, ...)]`, determining the ordering
        of monomial terms in the associated multivariable polynomial.
    """

    variables = int(variables)
    order = int(order)
    if variables < 1:
        raise ValueError('variables must be at least 1')
    if order < 1:
        raise ValueError('order must be at least 1')

    shape_details = tuple([order + 1 for _ in range(variables)])
    coefficient_list = []
    for index in numpy.ndindex(shape_details):
        total_exponent = sum(index)
        if total_exponent <= order:
            coefficient_list.append(index)
    return tuple(coefficient_list)


###################
# base rational polynomial fitting functions

def rational_poly_fit_1d(
        x: numpy.ndarray,
        data: numpy.ndarray,
        coeff_list: Sequence[Union[int, Tuple[int]]],
        cond: Optional[float] = None) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """
    Fits a one variable rational polynomial according to the input coefficient
    listing order.

    Parameters
    ----------
    x : numpy.ndarray
    data : numpy.ndarray
    coeff_list : List[int|Tuple[int]]
    cond : None|float
        Passed through to :func:`scipy.linalg.lstsq`.

    Returns
    -------
    numerator: numpy.ndarray
    denominator: numpy.ndarray

    Raises
    ------
    SarpyRatPolyError
        Convergence failures passed through
    """

    if coeff_list[0] not in [0, (0, )]:
        raise ValueError(
            'The first entry of coeff_list is required to be the constant term `0`')

    if not (x.size == data.size):
        raise ValueError('Size mismatch among data entries')
    x = x.flatten()
    data = data.flatten()

    # Enforcing that the denominator has constant term 1,
    #   P(x)/(1 + Q(x)) = d ->
    #   P(x) - d*Q(x) = d
    # This can be formulated as a strictly linear problem A*t = d

    A = numpy.empty((x.size, 2*len(coeff_list) - 1), dtype=numpy.float64)
    for i, entry in enumerate(coeff_list):
        if not (isinstance(entry, int) or (isinstance(entry, tuple) and len(entry) == 1 and isinstance(entry[0], int))):
            raise TypeError('coeff_list must be a list of integers or length 1 tuples of ints')
        if isinstance(entry, tuple):
            entry = entry[0]

        u = 1
        if entry > 0:
            u *= numpy.power(x, entry)
        A[:, i] = u
        if i > 0:
            A[:, i+len(coeff_list) - 1] = -u*data

    # perform least squares fit
    try:
        sol, residuals, rank, sing_values = lstsq(A, data, cond=cond)
    except LinAlgError as e:
        raise SarpyRatPolyError(str(e))

    # if len(residuals) != 0:
    residuals /= float(x.size)
    logger.info(
        'Performed rational polynomial fit, got\n\t'
        'residuals {}\n\t'
        'rank {}\n\t'
        'singular values {}'.format(residuals, rank, sing_values))

    numerator = numpy.zeros((len(coeff_list), ), dtype='float64')
    denominator = numpy.zeros((len(coeff_list), ), dtype='float64')
    denominator[0] = 1.0
    numerator[:] = sol[:len(coeff_list)]
    denominator[1:] = sol[len(coeff_list):]
    return numerator, denominator


def rational_poly_fit_2d(
        x: numpy.ndarray,
        y: numpy.ndarray,
        data: numpy.ndarray,
        coeff_list: Sequence[Tuple[int, int]],
        cond: Optional[float] = None) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """
    Fits a two variable rational polynomial according to the input coefficient
    listing order.

    Parameters
    ----------
    x : numpy.ndarray
    y : numpy.ndarray
    data : numpy.ndarray
    coeff_list : Sequence[Tuple[int, int]]
    cond : None|float
        Passed through to :func:`scipy.linalg.lstsq`.

    Returns
    -------
    numerator: numpy.ndarray
    denominator: numpy.ndarray

    Raises
    ------
    SarpyRatPolyError
        Convergence failures passed through
    """

    if coeff_list[0] != (0, 0):
        raise ValueError(
            'The first entry of coeff_list is required to be the constant term `(0, 0)`')

    if not (x.size == y.size and x.size == data.size):
        raise ValueError('Size mismatch among data entries')
    x = x.flatten()
    y = y.flatten()
    data = data.flatten()

    # Enforcing that the denominator has constant term 1,
    #   P(x, y)/(1 + Q(x, y)) = d ->
    #   P(x, y) - d*Q(x, y) = d
    # This can be formulated as a strictly linear problem A*t = d

    A = numpy.empty((x.size, 2 * len(coeff_list) - 1), dtype=numpy.float64)
    for i, entry in enumerate(coeff_list):
        if len(entry) != 2:
            raise TypeError('coeff_list must be a list of tuples of length 2')

        u = 1
        if entry[0] > 0:
            u *= numpy.power(x, entry[0])
        if entry[1] > 0:
            u *= numpy.power(y, entry[1])
        A[:, i] = u
        if i > 0:
            A[:, i + len(coeff_list) - 1] = -u*data

    # perform least squares fit
    try:
        sol, residuals, rank, sing_values = lstsq(A, data, cond=cond)
    except LinAlgError as e:
        raise SarpyRatPolyError(str(e))

    # if len(residuals) != 0:
    residuals /= float(x.size)
    logger.info(
        'Performed rational polynomial fit, got\n\t'
        'residuals {}\n\t'
        'rank {}\n\t'
        'singular values {}'.format(residuals, rank, sing_values))

    numerator = numpy.zeros((len(coeff_list),), dtype='float64')
    denominator = numpy.zeros((len(coeff_list),), dtype='float64')
    denominator[0] = 1.0
    numerator[:] = sol[:len(coeff_list)]
    denominator[1:] = sol[len(coeff_list):]
    return numerator, denominator


def rational_poly_fit_3d(
        x: numpy.ndarray,
        y: numpy.ndarray,
        z: numpy.ndarray,
        data: numpy.ndarray,
        coeff_list: Sequence[Tuple[int, int, int]],
        cond: Optional[float] = None) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """
    Fits a three variable rational polynomial according to the input coefficient
    listing order.

    Parameters
    ----------
    x : numpy.ndarray
    y : numpy.ndarray
    z : numpy.ndarray
    data : numpy.ndarray
    coeff_list : Sequence[Tuple[int, int, int]]
    cond : None|float
        Passed through to :func:`scipy.linalg.lstsq`.

    Returns
    -------
    numerator: numpy.ndarray
    denominator: numpy.ndarray

    Raises
    ------
    SarpyRatPolyError
        Convergence failures passed through
    """

    if coeff_list[0] != (0, 0, 0):
        raise ValueError(
            'The first entry of coeff_list is required to be the constant term `(0, 0, 0)`')

    if not (x.size == y.size and x.size == z.size and x.size == data.size):
        raise ValueError('Size mismatch among data entries')
    x = x.flatten()
    y = y.flatten()
    z = z.flatten()
    data = data.flatten()

    # Enforcing that the denominator has constant term 1,
    #   P(x, y, z)/(1 + Q(x, y, z)) = d ->
    #   P(x, y, z) - d*Q(x, y, z) = d
    # This can be formulated as a strictly linear problem A*t = d

    A = numpy.empty((x.size, 2*len(coeff_list) - 1), dtype=numpy.float64)
    for i, entry in enumerate(coeff_list):
        if len(entry) != 3:
            raise TypeError('coeff_list must be a list of tuples of length 3')

        u = 1
        if entry[0] > 0:
            u *= numpy.power(x, entry[0])
        if entry[1] > 0:
            u *= numpy.power(y, entry[1])
        if entry[2] > 0:
            u *= numpy.power(z, entry[2])
        A[:, i] = u
        if i > 0:
            A[:, i + len(coeff_list) - 1] = -u*data

    # perform least squares fit
    try:
        sol, residuals, rank, sing_values = lstsq(A, data, cond=cond)
    except LinAlgError as e:
        raise SarpyRatPolyError(str(e))

    # if len(residuals) != 0:
    residuals /= float(x.size)
    logger.info(
        'Performed rational polynomial fit, got\n\t'
        'residuals {}\n\t'
        'rank {}\n\t'
        'singular values {}'.format(residuals, rank, sing_values))

    numerator = numpy.zeros((len(coeff_list),), dtype='float64')
    denominator = numpy.zeros((len(coeff_list),), dtype='float64')
    denominator[0] = 1.0
    numerator[:] = sol[:len(coeff_list)]
    denominator[1:] = sol[len(coeff_list):]
    return numerator, denominator


####################
# rational polynomial definition

class RationalPolynomial(object):
    r"""
    A basic rational polynomial implementation. This assumes the data model
    `input_data -> output_data` via the relation

    .. math::
        X = (x, y, ...) & = (input\_data - input\_offset)/input\_scale \\
        (output\_data - output\_offset)/output\_scale & = Data = numerator(X)/denominator(X) \\
        output\_data & = (numerator(X)/denominator(X))*output_scale + output\_offset

    This object is callable, and acts as the evaluation function after construction.
    That is, suppose we have

    .. code::
        rational_poly = RationalPolynomial(...)  # suppose constructed as 2 variables
        output_0 = rational_poly([x, y])  # pass in the two variables as a single array
        output_1 = rational_poly(x, y)  # pass in the two variables individually
        # output_0 and output_1 should be identical

        output_fail = rational_poly(x, y, z)
        # this raises an exception for mismatch with the number of variables
    """

    __slots__ = (
        '_numerator', '_denominator', '_coeff_list', '_variables', '_input_offsets', '_input_scales',
        '_output_offset', '_output_scale', '_numerator_array', '_denominator_array')

    def __init__(
            self,
            numerator: Union[Sequence[float], numpy.ndarray],
            denominator: Union[Sequence[float], numpy.ndarray],
            coeff_list: Sequence[Tuple[int, ...]],
            input_offsets: Sequence[float],
            input_scales: Sequence[float],
            output_offset: float,
            output_scale: float):
        """

        Parameters
        ----------
        numerator : Sequence|numpy.ndarray
        denominator : Sequence|numpy.ndarray
        coeff_list : Sequence[Tuple[int, ...]]
        input_offsets : Sequence[float]
        input_scales : Sequence[float]
        output_offset : float
        output_scale : float
        """

        self._coeff_list = coeff_list
        self._variables = _get_num_variables(coeff_list)
        if self._variables not in [1, 2, 3]:
            raise ValueError('Functionality allows only 1, 2, or 3 variables.')

        if len(numerator) != len(self._coeff_list):
            raise ValueError('numerator must be the same length as coeff_list')
        self._numerator = numerator
        if len(denominator) != len(self._coeff_list):
            raise ValueError('denominator must be the same length as coeff_list')
        self._denominator = denominator

        if len(input_offsets) != self._variables:
            raise ValueError('The input_offsets must be the same length as the number of variables')
        self._input_offsets = input_offsets

        if len(input_scales) != self._variables:
            raise ValueError('The input_scale must be the same length as the number of variables')
        self._input_scales = input_scales

        self._output_offset = float(output_offset)
        self._output_scale = float(output_scale)

        self._numerator_array = _map_list_to_poly_matrix(numerator, coeff_list)
        self._denominator_array = _map_list_to_poly_matrix(denominator, coeff_list)

    @property
    def variables(self) -> int:
        """
        The number of independent variables.

        Returns
        -------
        int
        """

        return self._variables

    @property
    def coefficient_list(self) -> Sequence[Tuple[int, ...]]:
        """
        The coefficient list.

        Returns
        -------
        Sequence
        """

        return self._coeff_list

    @property
    def numerator(self) -> Sequence[float]:
        """
        The numerator coefficients.

        Returns
        -------
        Sequence
        """

        return self._numerator

    @property
    def denominator(self) -> Sequence[float]:
        """
        The denominator coefficients.

        Returns
        -------
        Sequence
        """

        return self._denominator

    def __call__(self, *input_variables: List[numpy.ndarray]) -> numpy.ndarray:
        def ensure_the_type(data):
            if isinstance(data, (numpy.number, int, float, numpy.ndarray)):
                return data
            else:
                return numpy.array(data)

        if len(input_variables) not in [1, self.variables]:
            raise ValueError('Got an unexpected number of input arguments')
        if len(input_variables) == 1:
            separate = False
            inp_vars = ensure_the_type(input_variables[0])
        else:
            separate = True
            inp_vars = [ensure_the_type(entry) for entry in input_variables]

        # todo: should we evaluate the viability of the input?
        if self.variables == 1:
            x = (inp_vars - self._input_offsets[0])/self._input_scales[0]
            value = polynomial.polyval(x, self._numerator_array) / \
                polynomial.polyval(x, self._denominator_array)
        elif self.variables == 2:
            if separate:
                x = (inp_vars[0] - self._input_offsets[0])/self._input_scales[0]
                y = (inp_vars[1] - self._input_offsets[1])/self._input_scales[1]
            else:
                if inp_vars.shape[-1] != 2:
                    raise ValueError(
                        'Final dimension of input data ({}) must match the number '
                        'of variables ({}).'.format(inp_vars.shape, self.variables))
                x = (inp_vars[..., 0] - self._input_offsets[0])/self._input_scales[0]
                y = (inp_vars[..., 1] - self._input_offsets[1])/self._input_scales[1]
            value = polynomial.polyval2d(x, y, self._numerator_array) / \
                polynomial.polyval2d(x, y, self._denominator_array)
        elif self.variables == 3:
            if separate:
                x = (inp_vars[0] - self._input_offsets[0])/self._input_scales[0]
                y = (inp_vars[1] - self._input_offsets[1])/self._input_scales[1]
                z = (inp_vars[2] - self._input_offsets[2])/self._input_scales[2]
            else:
                if inp_vars.shape[-1] != 3:
                    raise ValueError(
                        'Final dimension of input data ({}) must match the number '
                        'of variables ({}).'.format(inp_vars.shape, self.variables))
                x = (inp_vars[..., 0] - self._input_offsets[0])/self._input_scales[0]
                y = (inp_vars[..., 1] - self._input_offsets[1])/self._input_scales[1]
                z = (inp_vars[..., 2] - self._input_offsets[2]) / self._input_scales[2]
            value = polynomial.polyval3d(x, y, z, self._numerator_array) / \
                polynomial.polyval3d(x, y, z, self._denominator_array)
        else:
            raise ValueError('More than 3 variables is unsupported')
        return value*self._output_scale + self._output_offset


def _get_scale_and_offset(array: numpy.ndarray) -> Tuple[float, float]:
    min_value = numpy.min(array)
    max_value = numpy.max(array)
    scale_value = 0.5*(max_value - min_value)
    offset_value = 0.5*(max_value + min_value)
    return offset_value, scale_value


def get_rational_poly_1d(
        x: numpy.ndarray,
        data: numpy.ndarray,
        coeff_list: Optional[Sequence[Union[int, Tuple[int]]]] = None,
        order: Optional[int] = None,
        cond: Optional[float] = None) -> RationalPolynomial:
    """
    Gets the RationalPolynomial instance that comes from fitting the provided data.

    Parameters
    ----------
    x : numpy.ndarray
    data : numpy.ndarray
    coeff_list : None|Sequence
    order : None|int
    cond : None|float
        Passed through to :func:`scipy.linalg.lstsq`.

    Returns
    -------
    RationalPolynomial

    Raises
    ------
    SarpyRatPolyError
        Convergence failures passed through
    """

    if (coeff_list is None and order is None) or \
            (coeff_list is not None and order is not None):
        raise ValueError('Exact one of  coeff_list and order must be provided.')

    if order is not None:
        coeff_list = get_default_coefficient_ordering(1, int(order))

    if _get_num_variables(coeff_list) != 1:
        raise ValueError('The number of variables defined by the coefficient list must be 1.')

    scale_x, offset_x = _get_scale_and_offset(x)
    scale_data, offset_data = _get_scale_and_offset(data)

    numerator, denominator = rational_poly_fit_1d(
        (x-offset_x)/scale_x,
        (data-offset_data)/scale_data, coeff_list, cond=cond)

    return RationalPolynomial(
        numerator, denominator, coeff_list,
        (offset_x, ), (scale_x, ),
        offset_data, scale_data)


def get_rational_poly_2d(
        x: numpy.ndarray,
        y: numpy.ndarray,
        data: numpy.ndarray,
        coeff_list: Optional[Sequence[Tuple[int, int]]] = None,
        order: Optional[int] = None,
        cond: Optional[float] = None) -> RationalPolynomial:
    """
    Gets the RationalPolynomial instance that comes from fitting the provided data.

    Parameters
    ----------
    x : numpy.ndarray
    y : numpy.ndarray
    data : numpy.ndarray
    coeff_list : None|Sequence
    order : None|int
    cond : None|float
        Passed through to :func:`scipy.linalg.lstsq`.

    Returns
    -------
    RationalPolynomial

    Raises
    ------
    SarpyRatPolyError
        Convergence failures passed through
    """

    if (coeff_list is None and order is None) or \
            (coeff_list is not None and order is not None):
        raise ValueError('Exact one of coeff_list and order must be provided.')

    if order is not None:
        coeff_list = get_default_coefficient_ordering(2, int(order))

    if _get_num_variables(coeff_list) != 2:
        raise ValueError('The number of variables defined by the coefficient list must be 2.')

    scale_x, offset_x = _get_scale_and_offset(x)
    scale_y, offset_y = _get_scale_and_offset(y)
    scale_data, offset_data = _get_scale_and_offset(data)

    numerator, denominator = rational_poly_fit_2d(
        (x-offset_x)/scale_x, (y-offset_y)/scale_y,
        (data-offset_data)/scale_data, coeff_list, cond=cond)

    return RationalPolynomial(
        numerator, denominator, coeff_list,
        (offset_x, offset_y), (scale_x, scale_y),
        offset_data, scale_data)


def get_rational_poly_3d(
        x: numpy.ndarray,
        y: numpy.ndarray,
        z: numpy.ndarray,
        data: numpy.ndarray,
        coeff_list: Optional[Sequence[Tuple[int, int]]] = None,
        order: Optional[int] = None,
        cond: Optional[float] = None) -> RationalPolynomial:
    """
    Gets the RationalPolynomial instance that comes from fitting the provided data.

    Parameters
    ----------
    x : numpy.ndarray
    y : numpy.ndarray
    z : numpy.ndarray
    data : numpy.ndarray
    coeff_list : None|Sequence
    order : None|int
    cond : None|float
        Passed through to :func:`scipy.linalg.lstsq`.

    Returns
    -------
    RationalPolynomial

    Raises
    ------
    SarpyRatPolyError
        Convergence failures passed through
    """

    if (coeff_list is None and order is None) or \
            (coeff_list is not None and order is not None):
        raise ValueError('Exact one of coeff_list and order must be provided.')

    if order is not None:
        coeff_list = get_default_coefficient_ordering(3, int(order))

    if _get_num_variables(coeff_list) != 3:
        raise ValueError('The number of variables defined by the coefficient list must be 3.')

    scale_x, offset_x = _get_scale_and_offset(x)
    scale_y, offset_y = _get_scale_and_offset(y)
    scale_z, offset_z = _get_scale_and_offset(z)
    scale_data, offset_data = _get_scale_and_offset(data)

    numerator, denominator = rational_poly_fit_3d(
        (x-offset_x)/scale_x, (y-offset_y)/scale_y, (z-offset_z)/scale_z,
        (data-offset_data)/scale_data, coeff_list, cond=cond)

    return RationalPolynomial(
        numerator, denominator, coeff_list,
        (offset_x, offset_y, offset_z), (scale_x, scale_y, scale_z),
        offset_data, scale_data)


####################
# collective rational polynomial function

class CombinedRationalPolynomial(object):
    """
    Assemble a collection of RationalPolynomial objects with the same number of
    variables into a single multi-variable output object.
    """

    __slots__ = ('_collection', )

    def __init__(self, *collection: List[RationalPolynomial]):
        if len(collection) == 1 and isinstance(collection[0], Sequence):
            collection = collection[0]
        if len(collection) < 2:
            raise ValueError('This requires more than a single input')

        coll = []
        variables = None
        for entry in collection:
            if not isinstance(entry, RationalPolynomial):
                raise TypeError(
                    'Every input must be an instance of RationalPolynomial,\n\t'
                    'got type `{}`'.format(type(entry)))
            if variables is None:
                variables = entry.variables
            elif variables != entry.variables:
                raise TypeError(
                    'Every input must be an instance of RationalPolynomial with\n\t'
                    'the same number of variables, got type `{}` and `{}`'.format(variables, entry.variables))
            coll.append(entry)
        self._collection = tuple(coll)

    def __call__(
            self,
            *args: List[numpy.ndarray],
            combine: bool = True) -> Union[Tuple[numpy.ndarray, ...], numpy.ndarray]:
        out = tuple([entry(*args) for entry in self._collection])
        if combine:
            return numpy.stack(out, axis=-1)
        else:
            return out



"""
The error parameters type definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from typing import Union, Tuple, Optional

from sarpy.io.xml.base import Serializable
from sarpy.io.xml.descriptors import FloatDescriptor, SerializableDescriptor
from sarpy.io.complex.sicd_elements.blocks import ErrorDecorrFuncType
from sarpy.io.complex.sicd_elements.ErrorStatistics import PosVelErrType, TropoErrorType

from .base import DEFAULT_STRICT, FLOAT_FORMAT
from .blocks import AddedParametersType


class RadarSensorType(Serializable):
    """
    Radar sensor error statistics.
    """

    _fields = ('RangeBias', 'ClockFreqSF', 'CollectionStartTime', 'RangeBiasDecorr')
    _required = ('RangeBias', )
    _numeric_format = {'RangeBias': FLOAT_FORMAT, 'ClockFreqSF': FLOAT_FORMAT, 'CollectionStartTime': FLOAT_FORMAT}
    # descriptors
    RangeBias = FloatDescriptor(
        'RangeBias', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Range bias error standard deviation.')  # type: float
    ClockFreqSF = FloatDescriptor(
        'ClockFreqSF', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Payload clock frequency scale factor standard deviation, '
                  r'where :math:`SF = (\Delta f)/f_0`.')  # type: float
    CollectionStartTime = FloatDescriptor(
        'CollectionStartTime', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Collection Start time error standard deviation, '
                  'in seconds.')  # type: float
    RangeBiasDecorr = SerializableDescriptor(
        'RangeBiasDecorr', ErrorDecorrFuncType, _required, strict=DEFAULT_STRICT,
        docstring='Range Bias error decorrelation function.')  # type: ErrorDecorrFuncType

    def __init__(self, RangeBias=None, ClockFreqSF=None, CollectionStartTime=None,
                 RangeBiasDecorr=None, **kwargs):
        """

        Parameters
        ----------
        RangeBias : float
        ClockFreqSF : float
        CollectionStartTime : float
        RangeBiasDecorr : ErrorDecorrFuncType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.RangeBias = RangeBias
        self.ClockFreqSF = ClockFreqSF
        self.CollectionStartTime = CollectionStartTime
        self.RangeBiasDecorr = RangeBiasDecorr
        super(RadarSensorType, self).__init__(**kwargs)


class IonoErrorType(Serializable):
    """
    Ionosphere delay error statistics.
    """

    _fields = ('IonoRangeVertical', 'IonoRangeRateVertical', 'IonoRgRgRateCC', 'IonoRangeVertDecorr')
    _required = ('IonoRangeVertical', )
    _numeric_format = {'IonoRangeVertical': FLOAT_FORMAT, 'IonoRangeRateVertical': FLOAT_FORMAT,
                       'IonoRgRgRateCC': FLOAT_FORMAT}
    # descriptors
    IonoRangeVertical = FloatDescriptor(
        'IonoRangeVertical', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Ionosphere two-way delay error for normal incidence standard deviation. '
                  r'Expressed as a range error. :math:`(\Delta R) = (\Delta T) \cdot (c/2)`.')  # type: float
    IonoRangeRateVertical = FloatDescriptor(
        'IonoRangeRateVertical', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Ionosphere two-way delay rate of change error for normal incidence standard deviation. '
                  r'Expressed as a range rate error. :math:`\dot{R} = \Delta \dot{TD_Iono} \times c/2`.')  # type: float
    IonoRgRgRateCC = FloatDescriptor(
        'IonoRgRgRateCC', _required, strict=DEFAULT_STRICT, bounds=(-1, 1),
        docstring='Ionosphere range error and range rate error correlation coefficient.')  # type: float
    IonoRangeVertDecorr = SerializableDescriptor(
        'IonoRangeVertDecorr', ErrorDecorrFuncType, _required, strict=DEFAULT_STRICT,
        docstring='Ionosphere range error decorrelation function.')  # type: ErrorDecorrFuncType

    def __init__(self, IonoRangeVertical=None, IonoRangeRateVertical=None,
                 IonoRgRgRateCC=None, IonoRangeVertDecorr=None, **kwargs):
        """

        Parameters
        ----------
        IonoRangeVertical : float
        IonoRangeRateVertical : float
        IonoRgRgRateCC : float
        IonoRangeVertDecorr : ErrorDecorrFuncType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.IonoRangeVertical = IonoRangeVertical
        self.IonoRangeRateVertical = IonoRangeRateVertical
        self.IonoRgRgRateCC = IonoRgRgRateCC
        self.IonoRangeVertDecorr = IonoRangeVertDecorr
        super(IonoErrorType, self).__init__(**kwargs)


class BistaticRadarSensorType(Serializable):
    """
    Error statistics for a single radar platform.
    """

    _fields = ('DelayBias', 'ClockFreqSF', 'CollectionStartTime')
    _required = ('CollectionStartTime', )
    _numeric_format = {
        'DelayBias': FLOAT_FORMAT, 'ClockFreqSF': FLOAT_FORMAT,
        'CollectionStartTime': FLOAT_FORMAT}
    # descriptors
    DelayBias = FloatDescriptor(
        'DelayBias', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Optional[float]
    ClockFreqSF = FloatDescriptor(
        'ClockFreqSF', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Payload clock frequency scale factor standard deviation, '
                  r'where :math:`SF = (\Delta f)/f_0`.')  # type: Optional[float]
    CollectionStartTime = FloatDescriptor(
        'CollectionStartTime', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Collection Start time error standard deviation, '
                  'in seconds.')  # type: float

    def __init__(self, DelayBias=None, ClockFreqSF=None, CollectionStartTime=None, **kwargs):
        """

        Parameters
        ----------
        DelayBias : None|float
        ClockFreqSF : None|float
        CollectionStartTime : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.DelayBias = DelayBias
        self.ClockFreqSF = ClockFreqSF
        self.CollectionStartTime = CollectionStartTime
        super(BistaticRadarSensorType, self).__init__(**kwargs)

    def version_required(self) -> Tuple[int, int, int]:
        required = (1, 0, 1)
        if self.DelayBias is not None:
            required = max(required, (1, 1, 0))
        return required


class MonostaticType(Serializable):
    """
    Error parameters for monstatic collection.
    """

    _fields = ('PosVelErr', 'RadarSensor', 'TropoError', 'IonoError', 'AddedParameters')
    _required = ('PosVelErr', 'RadarSensor')
    # descriptors
    PosVelErr = SerializableDescriptor(
        'PosVelErr', PosVelErrType, _required, strict=DEFAULT_STRICT,
        docstring='Position and velocity error statistics for the sensor '
                  'platform.')  # type: PosVelErrType
    RadarSensor = SerializableDescriptor(
        'RadarSensor', RadarSensorType, _required, strict=DEFAULT_STRICT,
        docstring='Radar sensor error statistics.')  # type: RadarSensorType
    TropoError = SerializableDescriptor(
        'TropoError', TropoErrorType, _required, strict=DEFAULT_STRICT,
        docstring='Troposphere delay error statistics.')  # type: TropoErrorType
    IonoError = SerializableDescriptor(
        'IonoError', IonoErrorType, _required, strict=DEFAULT_STRICT,
        docstring='Ionosphere delay error statistics.')  # type: IonoErrorType
    AddedParameters = SerializableDescriptor(
        'AddedParameters', AddedParametersType, _required, strict=DEFAULT_STRICT,
        docstring='Additional error parameters that may be added')  # type: Union[None, AddedParametersType]

    def __init__(self, PosVelErr=None, RadarSensor=None, TropoError=None, IonoError=None,
                 AddedParameters=None, **kwargs):
        """

        Parameters
        ----------
        PosVelErr : PosVelErrType
        RadarSensor : RadarSensorType
        TropoError : None|TropoErrorType
        IonoError : None|IonoErrorType
        AddedParameters : None|AddedParametersType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.PosVelErr = PosVelErr
        self.RadarSensor = RadarSensor
        self.TropoError = TropoError
        self.IonoError = IonoError
        self.AddedParameters = AddedParameters
        super(MonostaticType, self).__init__(PosVelErr=PosVelErr, RadarSensor=RadarSensor, **kwargs)


class PlatformType(Serializable):
    """
    Basic bistatic platform error type definition.
    """
    _fields = ('PosVelErr', 'RadarSensor')
    _required = _fields
    # descriptors
    PosVelErr = SerializableDescriptor(
        'PosVelErr', PosVelErrType, _required, strict=DEFAULT_STRICT,
        docstring='Position and velocity error statistics for the sensor '
                  'platform.')  # type: PosVelErrType
    RadarSensor = SerializableDescriptor(
        'RadarSensor', BistaticRadarSensorType, _required, strict=DEFAULT_STRICT,
        docstring='Platform sensor error statistics.')  # type: BistaticRadarSensorType

    def __init__(self, PosVelErr=None, RadarSensor=None, **kwargs):
        """

        Parameters
        ----------
        PosVelErr : PosVelErrType
        RadarSensor : BistaticRadarSensorType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.PosVelErr = PosVelErr
        self.RadarSensor = RadarSensor
        super(PlatformType, self).__init__(**kwargs)

    def version_required(self) -> Tuple[int, int, int]:
        required = (1, 0, 1)
        if self.RadarSensor is not None:
            required = max(required, self.RadarSensor.version_required())
        return required


class BistaticType(Serializable):
    """
    Error parameters for bistatic parameters.
    """

    _fields = ('TxPlatform', 'RcvPlatform', 'AddedParameters')
    _required = ('TxPlatform', 'RcvPlatform')
    # descriptors
    TxPlatform = SerializableDescriptor(
        'TxPlatform', PlatformType, _required, strict=DEFAULT_STRICT,
        docstring='Error statistics for the transmit platform.')  # type: PlatformType
    RcvPlatform = SerializableDescriptor(
        'RcvPlatform', PlatformType, _required, strict=DEFAULT_STRICT,
        docstring='Error statistics for the receive platform.')  # type: PlatformType
    AddedParameters = SerializableDescriptor(
        'AddedParameters', AddedParametersType, _required, strict=DEFAULT_STRICT,
        docstring='Additional error parameters that may be added')  # type: Union[None, AddedParametersType]

    def __init__(self, TxPlatform=None, RcvPlatform=None, AddedParameters=None, **kwargs):
        """

        Parameters
        ----------
        TxPlatform : PlatformType
        RcvPlatform : PlatformType
        AddedParameters : None|AddedParametersType|dict
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.TxPlatform = TxPlatform
        self.RcvPlatform = RcvPlatform
        self.AddedParameters = AddedParameters
        super(BistaticType, self).__init__(**kwargs)

    def version_required(self) -> Tuple[int, int, int]:
        required = (1, 0, 1)
        for fld in ['TxPlatform', 'RcvPlatform']:
            val = getattr(self, fld)
            if val is not None:
                required = max(required, val.version_required())
        return required


class ErrorParametersType(Serializable):
    """
    Parameters that describe the statistics of errors in measured or estimated
    parameters that describe the collection.
    """

    _fields = ('Monostatic', 'Bistatic')
    _required = ()
    _choice = ({'required': True, 'collection': _fields}, )

    # descriptors
    Monostatic = SerializableDescriptor(
        'Monostatic', MonostaticType, _required, strict=DEFAULT_STRICT,
        docstring='The monstatic parameters.')  # type: Union[None, MonostaticType]
    Bistatic = SerializableDescriptor(
        'Bistatic', BistaticType, _required, strict=DEFAULT_STRICT,
        docstring='The bistatic parameters.')  # type: Union[None, BistaticType]

    def __init__(self, Monostatic=None, Bistatic=None, **kwargs):
        """

        Parameters
        ----------
        Monostatic : None|MonostaticType
        Bistatic : None|BistaticType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Monostatic = Monostatic
        self.Bistatic = Bistatic
        super(ErrorParametersType, self).__init__(**kwargs)

    def version_required(self) -> Tuple[int, int, int]:
        required = (1, 0, 1)
        if self.Bistatic is not None:
            required = max(required, self.Bistatic.version_required())
        return required




import numpy
from sarpy.geometry.geocoords import geodetic_to_ecf

from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class CMETAAType(TREElement):
    def __init__(self, value):
        super(CMETAAType, self).__init__()
        self.add_field('RELATED_TRES', 's', 2, value)
        self.add_field('ADDITIONAL_TRES', 's', 120, value)
        self.add_field('RD_PRC_NO', 's', 12, value)
        self.add_field('IF_PROCESS', 's', 4, value)
        self.add_field('RD_CEN_FREQ', 's', 4, value)
        self.add_field('RD_MODE', 's', 5, value)
        self.add_field('RD_PATCH_NO', 's', 4, value)
        self.add_field('CMPLX_DOMAIN', 's', 5, value)
        self.add_field('CMPLX_MAG_REMAP_TYPE', 's', 4, value)
        self.add_field('CMPLX_LIN_SCALE', 's', 7, value)
        self.add_field('CMPLX_AVG_POWER', 's', 7, value)
        self.add_field('CMPLX_LINLOG_TP', 's', 5, value)
        self.add_field('CMPLX_PHASE_QUANT_FLAG', 's', 3, value)
        self.add_field('CMPLX_PHASE_QUANT_BIT_DEPTH', 's', 2, value)
        self.add_field('CMPLX_SIZE_1', 's', 2, value)
        self.add_field('CMPLX_IC_1', 's', 2, value)
        self.add_field('CMPLX_SIZE_2', 's', 2, value)
        self.add_field('CMPLX_IC_2', 's', 2, value)
        self.add_field('CMPLX_IC_BPP', 's', 5, value)
        self.add_field('CMPLX_WEIGHT', 's', 3, value)
        self.add_field('CMPLX_AZ_SLL', 's', 2, value)
        self.add_field('CMPLX_RNG_SLL', 's', 2, value)
        self.add_field('CMPLX_AZ_TAY_NBAR', 's', 2, value)
        self.add_field('CMPLX_RNG_TAY_NBAR', 's', 2, value)
        self.add_field('CMPLX_WEIGHT_NORM', 's', 3, value)
        self.add_field('CMPLX_SIGNAL_PLANE', 's', 1, value)

        self.add_field('IF_DC_SF_ROW', 's', 6, value)
        self.add_field('IF_DC_SF_COL', 's', 6, value)

        for part in range(1, 5):
            for comp in ['ROW', 'COL']:
                self.add_field('IF_PATCH_{}{}'.format(comp, part), 's', 6, value)

        self.add_field('IF_DC_IS_ROW', 's', 8, value)
        self.add_field('IF_DC_IS_COL', 's', 8, value)
        self.add_field('IF_IMG_ROW_DC', 's', 8, value)
        self.add_field('IF_IMG_COL_DC', 's', 8, value)

        for part in range(1, 5):
            for comp in ['ROW', 'COL']:
                self.add_field('IF_TILE_{}{}'.format(comp, part), 's', 6, value)

        self.add_field('IF_RD', 's', 1, value)
        self.add_field('IF_RGWLK', 's', 1, value)
        self.add_field('IF_KEYSTN', 's', 1, value)
        self.add_field('IF_LINSFT', 's', 1, value)
        self.add_field('IF_SUBPATCH', 's', 1, value)
        self.add_field('IF_GEODIST', 's', 1, value)
        self.add_field('IF_RGFO', 's', 1, value)
        self.add_field('IF_BEAM_COMP', 's', 1, value)
        self.add_field('IF_RGRES', 's', 8, value)
        self.add_field('IF_AZRES', 's', 8, value)
        self.add_field('IF_RSS', 's', 8, value)
        self.add_field('IF_AZSS', 's', 8, value)
        self.add_field('IF_RSR', 's', 8, value)
        self.add_field('IF_AZSR', 's', 8, value)

        self.add_field('IF_RFFT_SAMP', 's', 7, value)
        self.add_field('IF_AZFFT_SAMP', 's', 7, value)
        self.add_field('IF_RFFT_TOT', 's', 7, value)
        self.add_field('IF_AZFFT_TOT', 's', 7, value)

        self.add_field('IF_SUBP_ROW', 's', 6, value)
        self.add_field('IF_SUBP_COL', 's', 6, value)
        self.add_field('IF_SUB_RG', 's', 4, value)
        self.add_field('IF_SUB_AZ', 's', 4, value)

        self.add_field('IF_RFFTS', 's', 1, value)
        self.add_field('IF_AFFTS', 's', 1, value)
        self.add_field('IF_RANGE_DATA', 's', 7, value)
        self.add_field('IF_INCPH', 's', 1, value)
        for part in range(1, 4):
            for comp in ['NAME', 'AMOUNT']:
                self.add_field('IF_SR_{}{}'.format(comp, part), 's', 8, value)
        for part in range(1, 4):
            self.add_field('AF_TYPE{}'.format(part), 's', 5, value)

        self.add_field('POL_TR', 's', 1, value)
        self.add_field('POL_RE', 's', 1, value)
        self.add_field('POL_REFERENCE', 's', 40, value)
        self.add_field('POL', 's', 1, value)
        self.add_field('POL_REG', 's', 1, value)
        self.add_field('POL_ISO_1', 's', 5, value)
        self.add_field('POL_BAL', 's', 1, value)
        self.add_field('POL_BAL_MAG', 's', 8, value)
        self.add_field('POL_BAL_PHS', 's', 8, value)
        self.add_field('POL_HCOMP', 's', 1, value)
        self.add_field('POL_HCOMP_BASIS', 's', 10, value)
        for part in range(1, 4):
            self.add_field('POL_HCOMP_COEF{}'.format(part), 's', 9, value)
        self.add_field('POL_AFCOMP', 's', 1, value)
        self.add_field('POL_SPARE_A', 's', 15, value)
        self.add_field('POL_SPARE_N', 's', 9, value)

        self.add_field('T_UTC_YYYYMMMDD', 's', 9, value)
        self.add_field('T_HHMMSSUTC', 's', 6, value)
        self.add_field('T_HHMMSSLOCAL', 's', 6, value)

        self.add_field('CG_SRAC', 's', 11, value)
        self.add_field('CG_SLANT_CONFIDENCE', 's', 7, value)
        self.add_field('CG_CROSS', 's', 11, value)
        self.add_field('CG_CROSS_CONFIDENCE', 's', 7, value)
        self.add_field('CG_CAAC', 's', 9, value)
        self.add_field('CG_CONE_CONFIDENCE', 's', 6, value)
        self.add_field('CG_GPSAC', 's', 8, value)
        self.add_field('CG_GPSAC_CONFIDENCE', 's', 6, value)
        self.add_field('CG_SQUINT', 's', 8, value)
        self.add_field('CG_GAAC', 's', 7, value)
        self.add_field('CG_GAAC_CONFIDENCE', 's', 6, value)
        self.add_field('CG_INCIDENT', 's', 7, value)
        self.add_field('CG_SLOPE', 's', 7, value)
        self.add_field('CG_TILT', 's', 8, value)
        self.add_field('CG_LD', 's', 1, value)
        self.add_field('CG_NORTH', 's', 8, value)
        self.add_field('CG_NORTH_CONFIDENCE', 's', 6, value)
        self.add_field('CG_EAST', 's', 8, value)
        self.add_field('CG_RLOS', 's', 8, value)
        self.add_field('CG_LOS_CONFIDENCE', 's', 6, value)
        self.add_field('CG_LAYOVER', 's', 8, value)
        self.add_field('CG_SHADOW', 's', 8, value)
        self.add_field('CG_OPM', 's', 7, value)
        self.add_field('CG_MODEL', 's', 5, value)

        for comp in ['X', 'Y', 'Z']:
            self.add_field('CG_AMPT_{}'.format(comp), 's', 13, value)
        self.add_field('CG_AP_CONF_XY', 's', 6, value)
        self.add_field('CG_AP_CONF_Z', 's', 6, value)

        for comp in ['X', 'Y', 'Z']:
            self.add_field('CG_APCEN_{}'.format(comp), 's', 13, value)
        self.add_field('CG_APER_CONF_XY', 's', 6, value)
        self.add_field('CG_APER_CONF_Z', 's', 6, value)

        for comp in ['X', 'Y', 'Z']:
            self.add_field('CG_FPNUV_{}'.format(comp), 's', 9, value)
        for comp in ['X', 'Y', 'Z']:
            self.add_field('CG_IDPNUV{}'.format(comp), 's', 9, value)
        for comp in ['X', 'Y', 'Z']:
            self.add_field('CG_SCECN_{}'.format(comp), 's', 13, value)
        self.add_field('CG_SC_CONF_XY', 's', 6, value)
        self.add_field('CG_SC_CONF_Z', 's', 6, value)
        self.add_field('CG_SWWD', 's', 8, value)
        for comp in ['X', 'Y', 'Z']:
            self.add_field('CG_SNVEL_{}'.format(comp), 's', 10, value)
        for comp in ['X', 'Y', 'Z']:
            self.add_field('CG_SNACC_{}'.format(comp), 's', 10, value)
        for comp in ['ROLL', 'PITCH', 'YAW']:
            self.add_field('CG_SNATT_{}'.format(comp), 's', 8, value)
        for comp in ['X', 'Y', 'Z']:
            self.add_field('CG_GTP_{}'.format(comp), 's', 9, value)

        self.add_field('CG_MAP_TYPE', 's', 4, value)
        if self.CG_MAP_TYPE.strip() == 'GEOD':
            self.add_field('CG_PATCH_LATCEN', 's', 11, value)
            self.add_field('CG_PATCH_LNGCEN', 's', 12, value)
            self.add_field('CG_PATCH_LTCORUL', 's', 11, value)
            self.add_field('CG_PATCH_LGCORUL', 's', 12, value)
            self.add_field('CG_PATCH_LTCORUR', 's', 11, value)
            self.add_field('CG_PATCH_LGCORUR', 's', 12, value)
            self.add_field('CG_PATCH_LTCORLR', 's', 11, value)
            self.add_field('CG_PATCH_LGCORLR', 's', 12, value)
            self.add_field('CG_PATCH_LTCORLL', 's', 11, value)
            self.add_field('CG_PATCH_LNGCOLL', 's', 12, value)
            self.add_field('CG_PATCH_LAT_CONFIDENCE', 's', 9, value)
            self.add_field('CG_PATCH_LONG_CONFIDENCE', 's', 9, value)
        elif self.CG_MAP_TYPE.strip() == 'MGRS':
            self.add_field('CG_MGRS_CENT', 's', 23, value)
            self.add_field('CG_MGRSCORUL', 's', 23, value)
            self.add_field('CG_MGRSCORUR', 's', 23, value)
            self.add_field('CG_MGRSCORLR', 's', 23, value)
            self.add_field('CG_MGRCORLL', 's', 23, value)
            self.add_field('CG_MGRS_CONFIDENCE', 's', 7, value)
            self.add_field('CG_MGRS_PAD', 's', 11, value)
        elif self.CG_MAP_TYPE.strip() == 'NA':
            self.add_field('CG_MAP_TYPE_BLANK', 's', 133, value)
        self.add_field('CG_SPARE_A', 's', 144, value)
        self.add_field('CA_CALPA', 's', 7, value)
        self.add_field('WF_SRTFR', 's', 14, value)
        self.add_field('WF_ENDFR', 's', 14, value)
        self.add_field('WF_CHRPRT', 's', 10, value)
        self.add_field('WF_WIDTH', 's', 9, value)
        self.add_field('WF_CENFRQ', 's', 13, value)
        self.add_field('WF_BW', 's', 13, value)
        self.add_field('WF_PRF', 's', 7, value)
        self.add_field('WF_PRI', 's', 9, value)
        self.add_field('WF_CDP', 's', 7, value)
        self.add_field('WF_NUMBER_OF_PULSES', 's', 9, value)
        self.add_field('VPH_COND', 's', 1, value)


class CMETAA(TREExtension):
    _tag_value = 'CMETAA'
    _data_type = CMETAAType

    def get_scp(self):
        """
        Gets the SCP location in ECF coordinates.

        Returns
        -------
        numpy.ndarray
        """

        cg_model = self.DATA.CG_MODEL.strip()
        scp_array = numpy.array(
            [float(self.DATA.CG_SCECN_X),
             float(self.DATA.CG_SCECN_Y),
             float(self.DATA.CG_SCECN_Z)], dtype='float64')
        if cg_model == 'ECEF':
            return scp_array
        elif cg_model == 'WGS84':
            return geodetic_to_ecf(scp_array)
        else:
            raise ValueError('Got unhandled CG_MODEL {}'.format(cg_model))

    def get_arp(self):
        """
        Gets the Aperture Position, at SCP Time, in ECF coordinates.

        Returns
        -------
        numpy.ndarray
        """

        cg_model = self.DATA.CG_MODEL.strip()
        arp_array = numpy.array(
            [float(self.DATA.CG_APCEN_X),
             float(self.DATA.CG_APCEN_Y),
             float(self.DATA.CG_APCEN_Z)], dtype='float64')
        if cg_model == 'ECEF':
            return arp_array
        elif cg_model == 'WGS84':
            return geodetic_to_ecf(arp_array)
        else:
            raise ValueError('Got unhandled CG_MODEL {}'.format(cg_model))

    def get_image_corners(self):
        """
        Gets the image corners in Lat/Lon, if feasible.

        Returns
        -------
        None|numpy.ndarray
        """

        if self.DATA.CG_MAP_TYPE.strip() != 'GEOD':
            return None

        return numpy.array([
                [float(self.DATA.CG_PATCH_LTCORUL), float(self.DATA.CG_PATCH_LGCORUL)],
                [float(self.DATA.CG_PATCH_LTCORUR), float(self.DATA.CG_PATCH_LGCORUR)],
                [float(self.DATA.CG_PATCH_LTCORLR), float(self.DATA.CG_PATCH_LGCORLR)],
                [float(self.DATA.CG_PATCH_LTCORLL), float(self.DATA.CG_PATCH_LNGCOLL)]], dtype='float64')




"""
The DataType definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = ("Thomas McCullough", "Michael Stewart, Valkyrie")

from typing import List

from sarpy.io.xml.base import Serializable
from sarpy.io.xml.descriptors import StringDescriptor, StringEnumDescriptor, \
    IntegerDescriptor, SerializableListDescriptor
from sarpy.io.phase_history.cphd1_elements.Data import SupportArraySizeType
from sarpy.io.phase_history.cphd1_elements.utils import binary_format_string_to_dtype

from .base import DEFAULT_STRICT


class ChannelSizeType(Serializable):
    """
    Parameters that define the Channel signal array and PVP array size and location.
    """

    _fields = ('Identifier', 'NumVectors', 'NumSamples', 'SignalArrayByteOffset', 'PVPArrayByteOffset')
    _required = _fields
    # descriptors
    Identifier = StringDescriptor(
        'Identifier', _required, strict=DEFAULT_STRICT,
        docstring='String that uniquely identifies the CRSD channel (Ch_ID)'
                  ' for which the data applies.')  # type: str
    NumVectors = IntegerDescriptor(
        'NumVectors', _required, strict=DEFAULT_STRICT, bounds=(1, None),
        docstring='Number of vectors in the signal array.')  # type: int
    NumSamples = IntegerDescriptor(
        'NumSamples', _required, strict=DEFAULT_STRICT, bounds=(1, None),
        docstring='Number of samples per vector in the signal array.')  # type: int
    SignalArrayByteOffset = IntegerDescriptor(
        'SignalArrayByteOffset', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Signal Array offset from the start of the Signal block (in bytes) to the '
                  'start of the Signal Array for the channel.')  # type: int
    PVPArrayByteOffset = IntegerDescriptor(
        'PVPArrayByteOffset', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='PVP Array offset from the start of the PVP block (in bytes) to the '
                  'start of the PVP Array for the channel.')  # type: int

    def __init__(self, Identifier=None, NumVectors=None, NumSamples=None, SignalArrayByteOffset=None,
                 PVPArrayByteOffset=None, **kwargs):
        """

        Parameters
        ----------
        Identifier : str
        NumVectors : int
        NumSamples : int
        SignalArrayByteOffset : int
        PVPArrayByteOffset : int
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Identifier = Identifier
        self.NumVectors = NumVectors
        self.NumSamples = NumSamples
        self.SignalArrayByteOffset = SignalArrayByteOffset
        self.PVPArrayByteOffset = PVPArrayByteOffset
        super(ChannelSizeType, self).__init__(**kwargs)


class DataType(Serializable):
    """
    Parameters that describe binary data components contained in the product.
    """

    _fields = (
        'SignalArrayFormat', 'NumBytesPVP', 'NumCRSDChannels',
        'Channels', 'NumSupportArrays', 'SupportArrays')
    _required = ('SignalArrayFormat', 'NumBytesPVP', 'Channels')
    _collections_tags = {
        'Channels': {'array': False, 'child_tag': 'Channel'},
        'SupportArrays': {'array': False, 'child_tag': 'SupportArray'}}
    # descriptors
    SignalArrayFormat = StringEnumDescriptor(
        'SignalArrayFormat', ('CI2', 'CI4', 'CF8'), _required, strict=DEFAULT_STRICT,
        docstring="Signal Array sample binary format of the CRSD signal arrays, where"
                  "`CI2` denotes a 1 byte signed integer parameter, 2's complement format, and 2 Bytes Per Sample;"
                  "`CI4` denotes a 2 byte signed integer parameter, 2's complement format, and 4 Bytes Per Sample;"
                  "`CF8` denotes a 4 byte floating point parameter, and 8 Bytes Per Sample.")  # type: str
    NumBytesPVP = IntegerDescriptor(
        'NumBytesPVP', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Number of bytes per set of Per Vector Parameters, where there is '
                  'one set of PVPs for each CRSD signal vector')  # type: int
    Channels = SerializableListDescriptor(
        'Channels', ChannelSizeType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that define the Channel signal array and PVP array size '
                  'and location.')  # type: List[ChannelSizeType]
    SupportArrays = SerializableListDescriptor(
        'SupportArrays', SupportArraySizeType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Support Array size parameters. Branch repeated for each binary support array. '
                  'Support Array referenced by its unique Support Array '
                  'identifier.')  # type: List[SupportArraySizeType]

    def __init__(self, SignalArrayFormat=None, NumBytesPVP=None, Channels=None, SupportArrays=None, **kwargs):
        """

        Parameters
        ----------
        SignalArrayFormat : str
        NumBytesPVP : int
        Channels : List[ChannelSizeType]
        SupportArrays : None|List[SupportArraySizeType]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.SignalArrayFormat = SignalArrayFormat
        self.NumBytesPVP = NumBytesPVP
        self.Channels = Channels
        self.SupportArrays = SupportArrays
        super(DataType, self).__init__(**kwargs)

    @property
    def NumSupportArrays(self):
        """
        int: The number of support arrays.
        """

        if self.SupportArrays is None:
            return 0
        else:
            return len(self.SupportArrays)

    @property
    def NumCRSDChannels(self):
        """
        int: The number of CRSD channels.
        """

        if self.Channels is None:
            return 0
        else:
            return len(self.Channels)

    def calculate_support_block_size(self):
        """
        Calculates the size of the support block in bytes as described by the SupportArray fields.
        """
        return sum([s.calculate_size() for s in self.SupportArrays])

    def calculate_pvp_block_size(self):
        """
        Calculates the size of the PVP block in bytes as described by the Data fields.
        """
        return self.NumBytesPVP * sum([c.NumVectors for c in self.Channels])

    def calculate_signal_block_size(self):
        """
        Calculates the size of the signal block in bytes as described by the Data fields.
        """
        num_bytes_per_sample = binary_format_string_to_dtype(self.SignalArrayFormat).itemsize
        return num_bytes_per_sample * sum([c.NumVectors * c.NumSamples for c in self.Channels])



"""
This package contains the AFRL RDE schema
"""

__classification__ = 'UNCLASSIFIED'

import pkg_resources


def get_schema_path(version='1.0.0'):
    """
    Location of AFRL/RDE schema file.

    Returns
    -------
    str
        The path to the ARFL/RDE schema.
    """

    if version == '1.0.0':
        return pkg_resources.resource_filename(
            'sarpy.annotation.afrl_rde_schema', 'afrl_rde_schema_v1.0.0_2022-02-15.xsd')
    else:
        raise ValueError('Got unrecognized version {}'.format(version))
